using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

/// <summary>
/// hide character version 4.0
/// a modification of Seamonkey's "hide character 1.0", which is a modification of Acidbubbles' "improved pov"
/// tl;dr: it adds a translucency and render queue slider for each part of the person's skin.
/// you can modify what parts are hidden by removing things under "MaterialsToHide" (the list starting with defaultMat)
/// but that's not really a problem anymore given that each part has a slider now lol
/// this edit was made by Doofenschmalphys
/// </summary>
public class HideCharacter : MVRScript
{
    private Atom _person;
    private DAZCharacterSelector _selector;
    private SkinHandler _skinHandler;
    private DAZCharacter _character;

    private bool _failedOnce = false;
    private bool disabled = false;
    private Queue<String> _error_details = new Queue<String>();
    private JSONStorableFloat transp;
    private byte iterations = 0;

    const String FILE_PATH = "Custom/Scripts/AnonymousCrow/TransparentPerson/ADD_ME.cs - ";

    public override void Init()
    {
        try
        {
            if (containingAtom?.type != "Person")
            {
                SuperController.LogError($"This plugin only works on Person atoms");
                DestroyImmediate(this);
                return;
            }

            _person = containingAtom;
            _selector = _person.GetComponentInChildren<DAZCharacterSelector>();

            Camera.onPreRender += OnPreRender;
            Camera.onPostRender += OnPreRender;
            transp = new JSONStorableFloat("Transparency", -1f, 1f, .0f, true);
        }
        catch (Exception e)
        {
            LogError(e);
            DestroyImmediate(this);
        }
    }

    private void OnPreRender(Camera cam)
    {

        try
        {
            if (_skinHandler != null)
                _skinHandler.BeforeRender();
        }
        catch (Exception e)
        {
            if (_failedOnce) return;
            _failedOnce = true;
            SuperController.LogError("something failed on prerender: " + e);
        }
    }

    private void OnPostRender(Camera cam)
    {

        try
        {
            if (_skinHandler != null)
                _skinHandler.AfterRender();
        }
        catch (Exception e)
        {
            if (_failedOnce) return;
            _failedOnce = true;
            SuperController.LogError("something failed on postrender: " + e);
        }
    }

    public void OnDisable()
    {
        try
        {
            ApplyAll(false);
            disabled = false;
            iterations = 0;
        }
        catch (Exception e)
        {
            LogError(e);
        }
    }

    public void OnDestroy()
    {
        OnDisable();
        Camera.onPreRender -= OnPreRender;
        Camera.onPostRender -= OnPostRender;
    }

    public void Update()
    {
        if (iterations > 32) return;

        try
        {
            Globals.counter = transp;
            if (Globals.newbie != null)
            {
                Globals.globalTrans.Add(Globals.newbie.material.name, new JSONStorableFloat(Globals.newbie.material.name, -1f, -1f, .0f, true));
                Globals.globalQueue.Add(Globals.newbie.material.name, new JSONStorableFloat(Globals.newbie.material.name + "-RenderQueue", (float)Globals.newbie.material.renderQueue, -1f, 5000f, true));
                CreateSlider(Globals.globalTrans[Globals.newbie.material.name], true);
                CreateSlider(Globals.globalQueue[Globals.newbie.material.name], false);
                //SuperController.LogError(Globals.newbie.material.shader.ToString());

                Globals.newbie = null;
            }
            ApplyAll(true);
        }
        catch (Exception e)
        {
            LogError(e);
        }
    }

    private void ApplyAll(bool active)
    {
        try
        {
            if (_selector.selectedCharacter?.skin == null) return;

            _character = _selector.selectedCharacter;

            if (UpdateHandler(ref _skinHandler, active))
                ConfigureHandler("Skin", ref _skinHandler, _skinHandler.Configure(_character.skin));
        }
        catch (Exception e)
        {
            e.Data.Add("function_name", "ApplyAll(bool active) #active = " + active);
            LogError(e);
            throw e;
        }
    }

    private void ConfigureHandler<T>(string what, ref T handler, int result) where T : IHandler, new()
    {
        switch (result)
        {
            case HandlerConfigurationResult.Success:
                iterations++;
                break;
            case HandlerConfigurationResult.CannotApply:
                handler = default(T);
                break;
            case HandlerConfigurationResult.TryAgainLater:
                handler = default(T);
                break;
        }
    }

    private bool UpdateHandler<T>(ref T handler, bool active)
     where T : IHandler, new()
    {
        if (handler == null && active)
        {
            handler = new T();
            return true;
        }

        if (handler != null && active)
        {
            handler.Restore();
            handler = new T();
            return true;
        }

        if (handler != null && !active)
        {
            handler.Restore();
            handler = default(T);
        }

        return false;
    }

    public static class HandlerConfigurationResult
    {
        public const int Success = 0;
        public const int CannotApply = 1;
        public const int TryAgainLater = 2;
    }

    public interface IHandler
    {
        void Restore();
        void BeforeRender();
        void AfterRender();
    }

    private void LogError(Exception e)
    {
        SuperController.LogError("Exception Message: " + e);
    }

    public class SkinHandler : IHandler
    {
        public class SkinShaderMaterialReference
        {
            public Material material;
            public Shader originalShader;
            public float originalAlphaAdjust;
            public float originalColorAlpha;
            public Color originalSpecColor;
			

            public static SkinShaderMaterialReference FromMaterial(Material material)
            {
                var materialRef = new SkinShaderMaterialReference();
                materialRef.material = material;
                materialRef.originalShader = material.shader;
                materialRef.originalAlphaAdjust = material.GetFloat("_AlphaAdjust");
                materialRef.originalColorAlpha = material.GetColor("_Color").a;
                materialRef.originalSpecColor = material.GetColor("_SpecColor");
		
                return materialRef;
            }
        }

        private static Dictionary<string, Shader> ReplacementShaders = new Dictionary<string, Shader>
        {
            // Opaque materials
            { "Custom/Subsurface/GlossCullComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossSeparateAlphaComputeBuff") },
            { "Custom/Subsurface/GlossNMCullComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
            { "Custom/Subsurface/GlossNMDetailCullComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMDetailNoCullSeparateAlphaComputeBuff") },
            { "Custom/Subsurface/CullComputeBuff", Shader.Find("Custom/Subsurface/TransparentSeparateAlphaComputeBuff") },
            { "Custom/Subsurface/GlossNMTessMappedFixedComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
            // Transparent materials
            { "Custom/Subsurface/TransparentGlossNoCullSeparateAlphaComputeBuff", null },
            { "Custom/Subsurface/TransparentGlossComputeBuff", null },
            { "Custom/Subsurface/TransparentComputeBuff", null },
            { "Custom/Subsurface/AlphaMaskComputeBuff", null },
			//{ "Custom/Subsurface/GlossNMTessMappedFixedComputeBuff", null },
            { "Marmoset/Transparent/Simple Glass/Specular IBLComputeBuff", null },
            { "Custom/Subsurface/TransparentGlossSeparateAlphaComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") }
        };

        private DAZSkinV2 _skin;
        private List<SkinShaderMaterialReference> _materialRefs;

        public int Configure(DAZSkinV2 skin)
        {
            _skin = skin;
            _materialRefs = new List<SkinShaderMaterialReference>();

            foreach (var material in skin.GPUmaterials)
            {
#if (IMPROVED_POV)
                if(material == null)
                    throw new InvalidOperationException("Attempts to apply the shader strategy on a destroyed material.");

                if (material.GetInt(SkinShaderMaterialReference.ImprovedPovEnabledShaderKey) == 1)
                    throw new InvalidOperationException("Attempts to apply the shader strategy on a skin that already has the plugin enabled (shader key).");
#endif

                var materialInfo = SkinShaderMaterialReference.FromMaterial(material);

                Shader shader;
                if (!ReplacementShaders.TryGetValue(material.shader.name, out shader))
                    SuperController.LogError("Missing replacement shader: '" + material.shader.name + "'");

                if (shader != null) material.shader = shader;

                //material.shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
                _materialRefs.Add(materialInfo);
            }

            // This is a hack to force a refresh of the shaders cache
            skin.BroadcastMessage("OnApplicationFocus", true);
            return HandlerConfigurationResult.Success;
        }

        public void Restore()
        {
            foreach (var material in _materialRefs)
                material.material.shader = material.originalShader;

            _materialRefs = null;

            // This is a hack to force a refresh of the shaders cache
            _skin.BroadcastMessage("OnApplicationFocus", true);
        }

        public void BeforeRender()
        {
            foreach (var materialRef in _materialRefs)
            {
                float t = Globals.counter.val;

                var material = materialRef.material;
                if (Globals.globalTrans.ContainsKey(material.name))
                {
                    t = Globals.globalTrans[material.name].val;
                    material.renderQueue = (int)Globals.globalQueue[material.name].val;

                    material.SetFloat("_AlphaAdjust", t);
                    var color = material.GetColor("_Color");
                    material.SetColor("_Color", new Color(color.r, color.g, color.b, 0f));
                    material.SetColor("_SpecColor", new Color(0f, 0f, 0f, 0f));
                }
                else
                {
                    Globals.newbie = materialRef;
                }
            }
        }

        public void AfterRender()
        {
            foreach (var materialRef in _materialRefs)
            {
                var material = materialRef.material;
                material.SetFloat("_AlphaAdjust", materialRef.originalAlphaAdjust);
                var color = material.GetColor("_Color");
                material.SetColor("_Color", new Color(color.r, color.g, color.b, materialRef.originalColorAlpha));
                material.SetColor("_SpecColor", materialRef.originalSpecColor);
            }
        }
    }
	
	static class Globals{
		public static JSONStorableFloat counter;
		public static SkinHandler.SkinShaderMaterialReference newbie = null;
		public static Dictionary<string, JSONStorableFloat> globalQueue = new Dictionary<string, JSONStorableFloat>{};
		public static Dictionary<string, JSONStorableFloat> globalTrans = new Dictionary<string, JSONStorableFloat>{};
		static int _getsetcounter;
		public static int getsetcounter
		{
			set{_getsetcounter=value;}
			get{return _getsetcounter;}
		}
        public static JSONStorableBool _hideLegs;
    }
}

